
/* Copyright 2021 Haakan T. Johansson */

/*  This file is part of WIGSGLL.
 *
 *  WIGSGLL is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  WIGSGLL is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with WIGSGLL.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */

#ifndef __WIGSGLL_H__
#define __WIGSGLL_H__

#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

/***********************************************************************
 *
 * Naming scheme:
 * 
 * wig <x> jj <calc-type> <iter> <return-type>
 *
 * x is 3, 6 or 9.
 * calc-type and return-type is d or qd for double or __float128.
 * iter is j or m, for the omitted parameter.
 * No iter for 9j, which only returns one symbol.
 */

/* Calculate using quad double (__float128), return as double.
 *
 * Recommended!
 */

int wig3jjqdjd  (/*   x   */ int two_j2, int two_j3,
                 /* -m2-m3*/ int two_m2, int two_m3,
                 double *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

int wig3jjqdmd  (int two_j1, int two_j2, int two_j3,
                 int two_m1, /*   x   */ /* -m1-x */
                 double *res, size_t entries,
                 int *two_m2_min, int *two_m2_max);

int wig6jjqdjd  (/*   x   */ int two_j2, int two_j3,
                 int two_l1, int two_l2, int two_l3,
                 double *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

int wig9jjqdd   (int two_j1, int two_j2, int two_j3,
		 int two_j4, int two_j5, int two_j6,
		 int two_j7, int two_j8, int two_j9,
		 double *res);

/* Calculate single symbols.
 *
 * Note: internally calculates string of symbols, so not faster!
 */

double wig3jjqdd(int two_j1, int two_j2, int two_j3,
		 int two_m1, int two_m2, int two_m3);

double wig6jjqdd(int two_j1, int two_j2, int two_j3,
		 int two_l1, int two_l2, int two_l3);

/* Calculate using quad double (__float128), return as double.
 *
 * Note that the result will not have full quad double precision
 * due to cancellation losses during calculations.
 *
 * Estimate: reached double precision around j ~ 10000 for 3j and 6j,
 * and around j ~ 30 for 9j.
 */

int wig3jjqdjqd (/*   x   */ int two_j2, int two_j3,
                 /* -m2-m3*/ int two_m2, int two_m3,
                 __float128 *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

int wig3jjqdmqd (int two_j1, int two_j2, int two_j3,
                 int two_m1, /*   x   */ /* -m1-x */
                 __float128 *res, size_t entries,
                 int *two_m2_min, int *two_m2_max);

int wig6jjqdjqd (/*   x   */ int two_j2, int two_j3,
                 int two_l1, int two_l2, int two_l3,
                 __float128 *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

int wig9jjqdqd  (int two_j1, int two_j2, int two_j3,
		 int two_j4, int two_j5, int two_j6,
		 int two_j7, int two_j8, int two_j9,
		 __float128 *res);

/* Calculate single symbols.  (Not faster!) */

__float128 wig3jjqdqd(int two_j1, int two_j2, int two_j3,
		      int two_m1, int two_m2, int two_m3);

__float128 wig6jjqdqd(int two_j1, int two_j2, int two_j3,
		      int two_l1, int two_l2, int two_l3);

/**********************************************************************/

/* From WIGXJPF: */

int trivial_zero_3j(int two_j1, int two_j2, int two_j3,
                    int two_m1, int two_m2, int two_m3);
int trivial_zero_6j(int two_j1, int two_j2, int two_j3,
                    int two_j4, int two_j5, int two_j6);
int trivial_zero_9j(int two_j1, int two_j2, int two_j3,
                    int two_j4, int two_j5, int two_j6,
                    int two_j7, int two_j8, int two_j9);

/**********************************************************************/

#ifdef __cplusplus
}
#endif

#endif/*__WIGSGLL_H__*/
