
/* Copyright 2023 Haakan T. Johansson */

/*  This file is part of WIGSGLL.
 *
 *  WIGSGLL is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  WIGSGLL is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with WIGSGLL.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */

#define DEBUG_WIGJJ 0

#if DEBUG_WIGJJ
# define IF_DEBUG_JJ(expr)  expr
#else
# define IF_DEBUG_JJ(expr)
#endif

#if 0
extern int interestingx;
#endif

/* Specialisations:
 *
 * Which symbol (3j or 6j), and which argument is iterated over:
 *
 * SGLL_ORDER        3   3   6
 * SGLL_ITER         j   m   j
 *
 * The above have set either SGLL_ORDER_3 or SGLL_ORDER_6,
 * and SGLL_ITER_j or SGLL_ITER_m.
 *
 * Floating point type used during calculations:
 *
 * CALC_TYPE         double   long double   __float128
 * CALC_SUFFIX       d        ld            qd
 *
 * Floating point type used to return results:
 *
 * RETURN_TYPE       double   long double   __float128
 * RETURN_SUFFIX     d        ld            qd
 */

/* Overall calculation plan:
 *
 * - Determine index limits.
 *   Sign of last value.
 *
 * - Calculate x[] and y[], (i.e. A and B, C and D, or E and F.
 *   These are calculated with integers until square root.
 *   They contain additional powers of two due the two_j notation.
 *
 * - Calculate s[] and r[] while absolute values stay < 1.
 *   This determines the limits of the three recursive ranges.
 *   simax and rimin are the respective last values, i.e. the
 *   ends of the non-classical regions.
 *
 * - Calculate un-normalised values.  In three ranges:
 *   [0 .. simax-1], [simax], [simax+1 .. rimin], [rimin+1 .. n-1]
 *   The value at [simax] is set as 1.
 *   During these calculations, also accumulate the norm.
 *
 * - Calculate the final norm, and rescale all result values.
 *
 * - Done!
 */

#define STRINGIFY_MACRO(x) STRINGIFY(x)
#define STRINGIFY(x) #x

/* Create the name of the function.
 * E.g wig6jjldjd, in pieces: wig 6 jj ld j d.
 * wig <order (3/6)> jj <calc-type> <iter-var (j/m)> <ret-type>
 */
#define FCN_POSTFIX2(order,iter,calc,ret)   wig##order##jj##calc##iter##ret
#define FCN_POSTFIX(order,iter,calc,ret)   FCN_POSTFIX2(order,iter,calc,ret)   

int FCN_POSTFIX(SGLL_ORDER,SGLL_ITER,CALC_SUFFIX,RETURN_SUFFIX)
     (
#if SGLL_ORDER_3
#if SGLL_ITER_j
      /*   X    */ int two_j2,  int two_j3,
      /* -m2-m3 */ int two_m2,  int two_m3,
#endif
#if SGLL_ITER_m
      int two_j1,  int two_j2,  int two_j3,
      int two_m1,  /*   X   */  /* -m1-X */
#endif
#endif
#if SGLL_ORDER_6
#if SGLL_ITER_j
      /*   X   */  int two_j2,  int two_j3,
      int two_l1,  int two_l2,  int two_l3,
#endif
#endif
      /* The result is always an array. */
      RETURN_TYPE *res,
      /* To avoid overflow, user shall tell size. */
      size_t entries,
#if SGLL_ITER_j
      int *two_j1_min_param, int *two_j1_max_param
#endif
#if SGLL_ITER_m
      int *two_m2_min_param, int *two_m2_max_param
#endif
      )
{
#if SGLL_ITER_j
  int two_j1_min, two_j1_max;
#endif
#if SGLL_ITER_m
  int two_m2_min, two_m2_max;
#endif
  int n;
  int i;
  int simax, rimin;

#define COLLECT_NEGATIVE(two_j1,two_j2,two_j3) do {     \
    collect_sign |= (two_j1) | (two_j2) | (two_j3);     \
  } while (0)

#define COLLECT_TRIANGLE_TRIVIAL_ZERO(two_j1,two_j2,two_j3) do {        \
    collect_odd  |= (two_j1) + (two_j2) + (two_j3);                     \
    collect_sign |= (((two_j2) + (two_j3)) - (two_j1));                 \
    collect_sign |= ((two_j1) - ((two_j2) - (two_j3)));                 \
    collect_sign |= ((two_j1) - ((two_j3) - (two_j2)));                 \
  } while (0)

#define COLLECT_ABS_M_WITHIN_J(two_m,two_j) do {        \
    collect_odd  |= ((two_m) + (two_j));                \
    collect_sign |= ((two_j) - (two_m));                \
    collect_sign |= ((two_j) + (two_m));                \
  } while (0)

  int collect_sign = 0;
  int collect_odd = 0;

#if SGLL_ORDER_6
  COLLECT_NEGATIVE(0     , two_j2, two_j3);
  COLLECT_NEGATIVE(two_l1, two_l2, two_l3);
  /*COLLECT_TRIANGLE_TRIVIAL_ZERO(two_j1, two_j2, two_j3);
    COLLECT_TRIANGLE_TRIVIAL_ZERO(two_j1, two_j5, two_j6);*/
  COLLECT_TRIANGLE_TRIVIAL_ZERO(two_l1, two_j2, two_l3);
  COLLECT_TRIANGLE_TRIVIAL_ZERO(two_l1, two_l2, two_j3);
#else
#if SGLL_ITER_j
  COLLECT_NEGATIVE(0     , two_j2, two_j3);

  COLLECT_ABS_M_WITHIN_J(two_m2, two_j2);
  COLLECT_ABS_M_WITHIN_J(two_m3, two_j3);
#endif
#if SGLL_ITER_m
  COLLECT_NEGATIVE(two_j1, two_j2, two_j3);
  COLLECT_TRIANGLE_TRIVIAL_ZERO(two_j1, two_j2, two_j3);

  COLLECT_ABS_M_WITHIN_J(two_m1, two_j1);
#endif
#endif

  if (((collect_sign & (1 << (sizeof (int) * 8 - 1))) |
       (collect_odd & 1)))
    {
      /* Set minimum and maximum parameters in case they are used as
       * limits to loop in calling code.
       */
#if SGLL_ITER_j
      if (two_j1_min_param)
	*two_j1_min_param = 0;
      if (two_j1_max_param)
	*two_j1_max_param = -1;
#endif
#if SGLL_ITER_m
      if (two_m2_min_param)
	*two_m2_min_param = 0;
      if (two_m2_max_param)
	*two_m2_max_param = -1;
#endif
      return 0;
    }

#if SGLL_ORDER_6
  two_j1_min = CHOOSE_MAX(abs(two_j2 - two_j3), abs(two_l2 - two_l3));
  two_j1_max = CHOOSE_MIN(    two_j2 + two_j3 ,     two_l2 + two_l3 );

  const int sign_i_max =
    ((two_j2 + two_j3 + two_l2 + two_l3) >> 1) & 1;

  n = (two_j1_max - two_j1_min) / 2 + 1;
#else
#if SGLL_ITER_j
  two_j1_min = CHOOSE_MAX(abs(two_j2 - two_j3), abs(two_m2 + two_m3));
  two_j1_max =                two_j2 + two_j3;

  const int sign_i_max =
    ((two_j2 - two_j3 + two_m2 + two_m3) >> 1) & 1;

  n = (two_j1_max - two_j1_min) / 2 + 1;
#endif
#if SGLL_ITER_m
  two_m2_min = CHOOSE_MAX(-two_j2, - two_j3 - two_m1);
  two_m2_max = CHOOSE_MIN( two_j2,   two_j3 - two_m1);
  
  const int sign_i_max =
    ((two_j2 - two_j3 - two_m1) >> 1) & 1;

  n = (two_m2_max - two_m2_min) / 2 + 1;
#endif
#endif

#if SGLL_ITER_j
  if (two_j1_min_param)
    *two_j1_min_param = two_j1_min;
  if (two_j1_max_param)
    *two_j1_max_param = two_j1_max;
#endif
#if SGLL_ITER_m
  if (two_m2_min_param)
    *two_m2_min_param = two_m2_min;
  if (two_m2_max_param)
    *two_m2_max_param = two_m2_max;
#endif

  if (n == 0)
    return 0;

  if (((size_t) n) > entries)
    {
      fprintf (stderr, "wigsgll: "
	       STRINGIFY_MACRO(SGLL_ORDER) "j-" STRINGIFY_MACRO(SGLL_ITER) " "
               "result has more symbols (%d) "
               "than result array (%d).\n",
	       n, (int) entries);
      wigsgll_error();
      return -1; /* Should not be reached, but anyhow. */
    }
  
#if DEBUG_WIGJJ
  printf ("-------------------------------------------------------\n");
#if SGLL_ORDER_6
  printf ("6j (%2d..%2d  %2d %2d ; %2d %2d %2d)\n",
	  two_j1_min, two_j1_max, two_j2, two_j3,
	  two_l1, two_l2, two_l3);
#endif
#if SGLL_ORDER_3
#if SGLL_ITER_j
  printf ("3j (%2d..%2d  %2d %2d ; %2d %2d %2d)\n",
	  two_j1_min, two_j1_max, two_j2, two_j3,
	  -two_m2-two_m3, two_m2, two_m3);
#endif
#if SGLL_ITER_m
  printf ("3j (%2d %2d %2d ; %2d %2d..%2d ...)\n",
	  two_j1, two_j2, two_j3,
	  two_m1, two_m2_min, two_m2_max);
#endif
#endif
#endif
  
  if (n == 1)
    {
#if SGLL_ORDER_6
      res[0] = (RETURN_TYPE) ((1 - 2 * sign_i_max) /
			      CALC_SQRT((two_j1_min + 1) * (two_l1 + 1)));
#endif
#if SGLL_ORDER_3
#if SGLL_ITER_j
      res[0] = (RETURN_TYPE) ((1 - 2 * sign_i_max) /
			      CALC_SQRT(two_j1_min + 1));
#endif
#if SGLL_ITER_m
      res[0] = (RETURN_TYPE) ((1 - 2 * sign_i_max) /
			      CALC_SQRT(two_j1 + 1));
#endif
#endif
      return 1;
    }

  CALC_TYPE x[n]; /* A, C, E */
  CALC_TYPE y[n]; /* B, D, F */

  int64_t jj = 0;
  int64_t jj_bias = 0;
  
#if SGLL_ORDER_3
#if SGLL_ITER_j
  const int64_t jj2 = two_j2;
  const int64_t jj3 = two_j3;
  const int64_t mm2 = two_m2;
  const int64_t mm3 = two_m3;

  const int64_t sqr_jj2_jj3   = SQR(jj2 - jj3);
  const int64_t sqr_jj2_jj3_2 = SQR(jj2 + jj3 + 2);

  const int64_t jj2_jj2_2 = jj2 * (jj2 + 2);
  const int64_t jj3_jj3_2 = jj3 * (jj3 + 2);

  const int64_t sqr_mm2_mm3   = SQR(mm2 + mm3);

  const int64_t fixed_B =
    (two_m2 + two_m3) * (jj2_jj2_2 - jj3_jj3_2);
  const int64_t lin_B = two_m2 - two_m3;
#endif
#if SGLL_ITER_m
  const int64_t jj1 = two_j1;
  const int64_t jj2 = two_j2;
  const int64_t jj3 = two_j3;
  const int64_t mm1 = two_m1;

  const int64_t jj1_jj1_2 = jj1 * (jj1 + 2);
  const int64_t jj2_jj2_2 = jj2 * (jj2 + 2);
  const int64_t jj3_jj3_2 = jj3 * (jj3 + 2);

  const int64_t jj2_2 = jj2 + 2;
  const int64_t jj3_m_mm1_2 = jj3 - mm1 + 2;
  const int64_t jj3_p_mm1   = jj3 + mm1;

  const int64_t lin_D = jj2_jj2_2 + jj3_jj3_2 - jj1_jj1_2;
#endif
#endif

#if SGLL_ORDER_6
  const int64_t jj2 = two_j2;
  const int64_t jj3 = two_j3;
  const int64_t ll1 = two_l1;
  const int64_t ll2 = two_l2;
  const int64_t ll3 = two_l3;

  const int64_t sqr_jj2_jj3   = SQR(jj2 - jj3);
  const int64_t sqr_jj2_jj3_2 = SQR(jj2 + jj3 + 2);

  const int64_t sqr_ll2_ll3   = SQR(ll2 - ll3);
  const int64_t sqr_ll2_ll3_2 = SQR(ll2 + ll3 + 2);

  const int64_t jj2_jj2_2 = jj2 * (jj2 + 2);
  const int64_t jj3_jj3_2 = jj3 * (jj3 + 2);
  const int64_t ll1_ll1_2 = ll1 * (ll1 + 2);
  const int64_t ll2_ll2_2 = ll2 * (ll2 + 2);
  const int64_t ll3_ll3_2 = ll3 * (ll3 + 2);

  const int64_t fixed_F =
      ll2_ll2_2 * ( + jj2_jj2_2 - jj3_jj3_2)
    + ll3_ll3_2 * ( - jj2_jj2_2 + jj3_jj3_2);
  const int64_t lin_F =
    jj2_jj2_2 + jj3_jj3_2 - 2*ll1_ll1_2 + ll2_ll2_2 + ll3_ll3_2;

  (void) ll3_ll3_2;
  (void) jj3_jj3_2;
  (void) sqr_ll2_ll3_2;
  (void) sqr_jj2_jj3_2;
  (void) sqr_ll2_ll3;
  (void) sqr_jj2_jj3;
  (void) fixed_F;
  (void) lin_F;
#endif
  
  for (int i = 0; i < n; i++)
    {
#if SGLL_ITER_j
      int two_j1 = two_j1_min + 2 * i;
      int64_t jj = two_j1;
      const int64_t sqr_jj  = SQR(jj);
      const int64_t jj_jj_2 = jj * (jj + 2);
#endif
#if SGLL_ITER_m
      int two_m2 = two_m2_min + 2 * i;
      int64_t mm = two_m2;
#endif
      
#if SGLL_ORDER_6
      /*
      x[i] = CALC_SQRT((SQR(two_j1) - SQR(two_j2 - two_j3)) *
		       (SQR(two_j2 + two_j2 + 2) - SQR(two_j1)) *
		       (SQR(two_j1) - SQR(two_l2 - two_l3)) *
		       (SQR(two_l2 + two_l2 + 2) - SQR(two_j1)) );
 
      y[i] = (2 * two_j1 + 2) *
	(two_j1 * (two_j1 + 2) * (-two_j1 * (two_j1 + 2) +
				  two_j2 * (two_j2 + 2) +
				  two_j3 * (two_j3 + 2) +
				  -2 * two_l1 * (two_l1 + 2)) +
	 two_l2 * (two_l2 + 2) * (two_j1 * (two_j1 + 2) +
				  two_j2 * (two_j2 + 2) -
				  two_j3 * (two_j3 + 2)) +
	 two_l3 * (two_l3 + 2) * (two_j1 * (two_j1 + 2) -
				  two_j2 * (two_j2 + 2) +
				  two_j3 * (two_j3 + 2)) );
      */
      x[i] = CALC_SQRT(((CALC_TYPE)((sqr_jj         - sqr_jj2_jj3) *
				    (sqr_jj2_jj3_2  - sqr_jj     ))) *
		       ((CALC_TYPE)((sqr_jj         - sqr_ll2_ll3) *
				    (sqr_ll2_ll3_2  - sqr_jj     ))));
      y[i] = (((CALC_TYPE) (2*jj + 2) *
	       (CALC_TYPE) (jj_jj_2 * (-jj_jj_2 + lin_F) + fixed_F)));
#endif
#if SGLL_ORDER_3
#if SGLL_ITER_j
      x[i] = CALC_SQRT(((CALC_TYPE)((sqr_jj         - sqr_jj2_jj3) *
                                    (sqr_jj2_jj3_2  - sqr_jj     ))) *
		       ((CALC_TYPE)((sqr_jj         - sqr_mm2_mm3))));
      /*
      y[i] = (2 * two_j1 + 2) *
	((two_m2 + two_m3) * (jj2_jj2_2 - jj3_jj3_2) -
	 (two_m2 - two_m3) * jj_jj_2);
      */
      y[i] = (CALC_TYPE) ((2 * jj + 2) *
			  (fixed_B - lin_B * jj_jj_2));
#endif
#if SGLL_ITER_m
      /*
      x[i] = CALC_SQRT((CALC_TYPE)((two_j2 - two_m2 + 2) *
				   (two_j2 + two_m2) *
				   (two_j3 - two_m2 - two_m1 + 2) *
				   (two_j3 + two_m2 + two_m1)));
      y[i] = jj2_jj2_2 + jj3_jj3_2 - jj1_jj1_2 -
	2 * mm * (mm + two_m1);
      */
      x[i] = CALC_SQRT((CALC_TYPE)((jj2_2       - mm) *
				   (jj2         + mm) *
				   (jj3_m_mm1_2 - mm) *
				   (jj3_p_mm1   + mm)));
      y[i] = (CALC_TYPE) (lin_D -
			  2 * mm * (mm + two_m1));
#endif
#endif
      (void) two_j1; /* until 3j also */
    }

  CALC_TYPE r[n+1];
  CALC_TYPE s[n+1];

  IF_DEBUG_JJ(int setr[n+1]);
  IF_DEBUG_JJ(int sets[n+1]);
  IF_DEBUG_JJ(char calcby[n]);

  /* Intermediate result, with calculation precision. */
  CALC_TYPE res2[n+1];
  /* Intermediate from previous iteration. */
  CALC_TYPE res2_p;
  CALC_TYPE res2_pp;
  CALC_TYPE res2_pp_use;

  /* Summed weighted norm of the intermediate results. */
  CALC_TYPE norm_sqr;
  
#if DEBUG_WIGJJ
  for (i = 0; i < n+1; i++)
    {
      IF_DEBUG_JJ(setr[i] = 0);
      IF_DEBUG_JJ(sets[i] = 0);
    }
#endif
  
#if SGLL_ITER_j
  if (two_j1_min == 0)
    {
#if SGLL_ORDER_6
      /* two_j1_min = 0 => two_j2 = two_j3 and two_l2 = two_l3 */
      s[0] = -x[1] / ((CALC_TYPE) (jj2_jj2_2 + ll2_ll2_2 - ll1_ll1_2)) / 8.;
#endif
#if SGLL_ORDER_3
      /* two_j1_min = 0 => two_j2 = two_j3 and two_m2 = -two_m3 */
      /*
      printf ("%.6f %.6f %.6f XXX\n",
	      (double) CALC_SQRT(((CALC_TYPE)((sqr_jj2_jj3_2  - 0)))),
	      (double) ((2) * ( - (2 * two_m2))),
	      (double) x[1]);
      */
      s[0] = x[1] / two_m2 / 8.;
#endif
      IF_DEBUG_JJ(sets[0] = 1);
    }
  else
    {
      s[0] = -two_j1_min * x[1] / y[0];
      IF_DEBUG_JJ(sets[0] = 1);
    }
#endif

#if SGLL_ITER_m
  s[0] = - x[1] / y[0];
  IF_DEBUG_JJ(sets[0] = 1);
#endif

  i = 0;
  simax = -1;
  for ( ; ; )
    {
#if SGLL_ITER_j
      int two_j1;
#endif

      if (CALC_FABS(s[i]) > 1)
	break;
      simax = i;
      i++;
      if (i >= n-1)
	break;

#if SGLL_ITER_j
      two_j1 = two_j1_min + 2 * i;
#endif

#if SGLL_ITER_j
      s[i] = - two_j1 * x[i+1] / (y[i] + (two_j1 + 2) * x[i] * s[i-1]);
      IF_DEBUG_JJ(sets[i] = 1);
#endif
#if SGLL_ITER_m
      s[i] = - x[i+1] / (y[i] + x[i] * s[i-1]);
      IF_DEBUG_JJ(sets[i] = 1);
#endif
    }

#if SGLL_ITER_j
  r[n-1] = -(two_j1_max + 2) * x[n-1] / y[n-1];
  IF_DEBUG_JJ(setr[n-1] = 1);
#endif
#if SGLL_ITER_m
  r[n-1] = -x[n-1] / y[n-1];
  IF_DEBUG_JJ(setr[n-1] = 1);
#endif

  i = n-1;
  rimin = n+1;  /* n ??????? */
  for ( ; ; )
    {
#if SGLL_ITER_j
      int two_j1;
#endif

      if (CALC_FABS(r[i]) > 1)
	break;
      rimin = i;
      i--;
      if (i <= simax || i <= 0)
	break;

#if SGLL_ITER_j
      two_j1 = two_j1_min + 2 * i;
#endif

#if SGLL_ITER_j
      r[i] = -(two_j1 + 2) * x[i] / (y[i] + two_j1 * x[i+1] * r[i+1]);
      IF_DEBUG_JJ(setr[i] = 1);
#endif
#if SGLL_ITER_m
      r[i] = -x[i] / (y[i] + x[i+1] * r[i+1]);
      IF_DEBUG_JJ(setr[i] = 1);
#endif
    }

  /* The basis of the result is [simax+1], which is set to 1.  Smaller
   * [] are evaluated using s.  [] larger than rimin are evaluated
   * using r.  Between simax and rimin, Schulten-Gordon is used.
   */

#if DEBUG_WIGJJ  
#define DEBUG_NORM(c,i,norm,bias,res) \
  printf("%c %2d %30.20g %30.20g %30.20g\n", \
	 c, i, (double) norm_sqr, (double) bias, (double) res2_p)
#else
#define DEBUG_NORM(c,i,norm,bias,res) do { } while(0)
#endif
  
  /* Prepare backward iteration. */
  i = simax;
#if SGLL_ITER_j
  jj = two_j1_min + 2 * i;
  jj_bias = 2*jj + 2;
#endif
  
  /* Value [simax + 1]. */
  res2_p = res2[simax + 1] = 1.;
#if SGLL_ITER_j
  norm_sqr = (CALC_TYPE) (jj_bias + 4); /* * 1. */
#endif
#if SGLL_ITER_m
  norm_sqr = 1.;
#endif
  DEBUG_NORM('1', i, norm_sqr, jj_bias + 4, 1.);
  IF_DEBUG_JJ(calcby[simax + 1] = '1');
  
  /* Values [0 .. simax]. */  
  for ( ; i >= 0; i--, jj_bias -= 4)
    {
      IF_DEBUG_JJ(calcby[i] = '<');
      res2_p = res2[i] = s[i] * res2_p;
#if SGLL_ITER_j
      norm_sqr += ((CALC_TYPE) jj_bias) * SQR(res2_p);
#endif
#if SGLL_ITER_m
      norm_sqr += SQR(res2_p);
#endif
      DEBUG_NORM('<', i, norm_sqr, jj_bias, res2_p);
    }

  /* Prepare forward iteration (classical region).*/
  i = simax + 1;
#if SGLL_ITER_j
  jj = two_j1_min + 2 * i;
  jj_bias = 2*jj + 2 + 4;
#endif
  
  /* If we start at the bottom, the second value is gotten
   * with s[0] instead of the three-term formula.
   *
   * Is the same thing, alternative would be to set res2_pp[-1] = 0;
   */
  if (i == 0)
    {
      IF_DEBUG_JJ(calcby[1] = '=');
      res2_pp = 1.;
      res2_p = res2[1] = 1. / s[0];
#if SGLL_ITER_j
      norm_sqr += ((CALC_TYPE) jj_bias) * SQR(res2_p);
#endif
#if SGLL_ITER_m
      norm_sqr += SQR(res2_p);
#endif
      DEBUG_NORM('=', i, norm_sqr, jj_bias, res2_p);
      i++;
      jj += 2;
      jj_bias += 4;
    }
  else
    {
      res2_p = res2[i];
      res2_pp = res2[i-1];
    }

  /* Values [simax + 2 .. rimin]. */
  for ( ; i < rimin-1 && i < n-1; i++, jj += 2, jj_bias += 4)
    {
      IF_DEBUG_JJ(calcby[i+1] = '-');
      res2_pp_use = res2_pp;
      res2_pp = res2_p;
#if SGLL_ITER_j
      res2_p = res2[i+1] =
        -(y[i] * res2_p + ((CALC_TYPE) (jj + 2)) * x[i] * res2_pp_use) /
	(((CALC_TYPE) jj) * x[i+1]);
#endif
#if SGLL_ITER_m
      res2_p = res2[i+1] =
        -(y[i] * res2_p + x[i] * res2_pp_use) / (x[i+1]);
#endif
#if SGLL_ITER_j
      norm_sqr += ((CALC_TYPE) jj_bias) * SQR(res2_p);
#endif
#if SGLL_ITER_m
      norm_sqr += SQR(res2_p);
#endif
      DEBUG_NORM('-', i, norm_sqr, jj_bias, res2_p);
    }

  /* Values [rimin + 1 .. n-1]. */
  for ( ; i < n-1; i++, jj_bias += 4)
    {
      IF_DEBUG_JJ(calcby[i+1] = '>');
      res2_p = res2[i+1] = r[i+1] * res2_p;
#if SGLL_ITER_j
      norm_sqr += ((CALC_TYPE) jj_bias) * SQR(res2_p);
#endif
#if SGLL_ITER_m
      norm_sqr += SQR(res2_p);
#endif
      DEBUG_NORM('>', i, norm_sqr, jj_bias, res2_p);
    }

#if SGLL_ORDER_6
  norm_sqr *= 0.25 * (CALC_TYPE) (2 * ll1 + 2);
#endif
#if SGLL_ORDER_3
#if SGLL_ITER_j
  norm_sqr *= 0.5;
#endif
#if SGLL_ITER_m
  norm_sqr *= 0.5 * (CALC_TYPE) (2 * jj1 + 2);
#endif
#endif

  /* Determine the scale factor to reach the correct normalised result. */
  /* Note: signbit returns non-zero, not necessarily 1 for negative. */
  
  CALC_TYPE mult =
    (1 - 2 * (sign_i_max ^ !!signbit(res2[n-1]))) / CALC_SQRT(norm_sqr);

#if DEBUG_WIGJJ
  printf("%30.20g %30.20g %d %d\n",
	 (double) norm_sqr, (double) mult,
	 sign_i_max, signbit(res2[n-1]));
#endif
  
  /* Rescale the results. */  
  for (i = 0; i < n; i++)
    {
      res[i] = (RETURN_TYPE) (res2[i] * mult);
#if DEBUG_WIGJJ
      printf("%30.20g %30.20g\n", (double) res[i], (double) res2[i]);
#endif
    }

#if DEBUG_WIGJJ
  if (1)
    {
      for (i = 0; i < n; ++i)
        {
          printf ("[%2d] %2d: ",
                  i,
#if SGLL_ITER_j
		  two_j1_min + i * 2
#endif
#if SGLL_ITER_m
		  two_m2_min + i * 2
#endif
		  );
          printf ("%15.4f %15.4f ",
#if SGLL_ORDER_6
                  (double) x[i]/16.,
                  (double) y[i]/32.
#endif
#if SGLL_ORDER_3
#if SGLL_ITER_j
                  (double) x[i]/8.,
                  (double) y[i]/16.
#endif
#if SGLL_ITER_m
                  (double) x[i]/4.,
                  (double) y[i]/4.
#endif
#endif
		  );
          if (sets[i])
            printf (" %10.4g", (double) s[i]);
          else
            printf (" %10s", "-");
          if (setr[i])
            printf (" %10.4g", (double) r[i]);
          else
            printf (" %10s", "-");
	  printf (" %10.4f", (double) res2[i]);

	  printf ("  %c %c",
		  i < simax+1 ? '<' :
		  (i == simax+1 ? '1' :
		   (i <= rimin-1 ? '-' : '>')),
		  calcby[i]);
	  
	  /*
          printf ("  %c%c%c",
                  iminus == i ? '-' : ' ',
                  iplus  == i ? '+' : ' ',
                  istart == i ? 's' : ' ');
	  */
	  printf ("\n");
        }
    }
  /*
  if (fabs(res1[0]) < 1e-7)
    printf ("BORDER- 0\n");
  if (fabs(res1[nj-1]) < 1e-7)
    printf ("BORDER+ 0\n");
  */
#endif

#if 0
  if (n > 100 && n < 200 &&
      simax > 10 && simax < 25 &&
      (n-rimin) > 10)
    {
      for (i = 0; i < simax; i++)
	if (s[i] < 0)
	  goto notthis;
      for (i = rimin+1; i < n; i++)
	if (r[i] < 0)
	  goto notthis;

      int signflips = 0;

      for (i = 1; i < n; i++)
	{
	  if (res[i] * res[i-1] < 0)
	    signflips++;
	}

      if (signflips > n/10 ||
	  signflips < n/20)
	goto notthis;

      interestingx = 1;
      printf ("%3d %3d %3d %d\n",
	      n, simax, n-rimin, signflips);
#if SGLL_ORDER_6
      printf ("6j (%2d..%2d  %2d %2d ; %2d %2d %2d)\n",
              two_j1_min, two_j1_max, two_j2, two_j3,
              two_l1, two_l2, two_l3);
#endif
    notthis:;
    }
#endif

  return n;
}

#undef IF_DEBUG_JJ
